/*
 * Generic receive layer for the SA1100 USB client function
 * Copyright (c) 2001 by Nicolas Pitre
 * 
 * This code was loosely inspired by the original version which was
 * Copyright (c) Compaq Computer Corporation, 1998-1999
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This is still work in progress...
 * 
 */

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <asm/dma.h>
#include <asm/system.h>

#include "usb_ctl.h"


static char *ep1_buf;
static int ep1_len;
static usb_callback_t ep1_callback;
static char *ep1_curdmabuf;
static dma_addr_t ep1_curdmapos;
static int ep1_curdmalen;
static int ep1_remain;
static int ep1_stall;

static int dmachn_rx;


static void
ep1_start(void)
{
	if (!ep1_len)
		return;
	
	sa1100_dma_flush_all(dmachn_rx);
	if (!ep1_curdmalen) {
		ep1_curdmalen = usbd_info.rx_pktsize;
		if (ep1_curdmalen > ep1_remain)
			ep1_curdmalen = ep1_remain;
		ep1_curdmapos = pci_map_single(NULL, ep1_curdmabuf, ep1_curdmalen, 
					       PCI_DMA_FROMDEVICE);
	}

	//UDC_write(Ser0UDCOMP, ep1_curdmalen-1);
	sa1100_dma_queue_buffer(dmachn_rx, NULL, ep1_curdmapos, ep1_curdmalen);
}

static void
ep1_done(int flag)
{
	int size = ep1_len - ep1_remain;
	if (!ep1_len)
		return;
	if (ep1_curdmalen)
		pci_unmap_single(NULL, ep1_curdmapos, ep1_curdmalen, 
				 PCI_DMA_FROMDEVICE);
	ep1_len = ep1_curdmalen = 0;
	if (ep1_callback)
		ep1_callback(flag, size);
}

int 
ep1_init(int chn)
{
	dmachn_rx = chn;
	sa1100_dma_flush_all(dmachn_rx);
#if 1
	ep1_stall = 1;
        UDC_set(Ser0UDCCS1, UDCCS1_FST);
        UDC_clear(Ser0UDCCS1, UDCCS1_FST);
#endif
	ep1_done(-EAGAIN);
	return 0;
}

void
ep1_reset(void)
{
	sa1100_dma_flush_all(dmachn_rx);
#if 1
	ep1_stall = 1;
        UDC_set(Ser0UDCCS1, UDCCS1_FST);
        UDC_clear(Ser0UDCCS1, UDCCS1_FST);
#endif
	ep1_done(-EINTR);
}

void 
ep1_int_hndlr(int udcsr)
{
	dma_addr_t dma_addr;
	unsigned int len;
	int status = Ser0UDCCS1;

	if (status & UDCCS1_RPC) {
		
		if (status & UDCCS1_SST) {
			//printk("usb_recv: stall sent\n");
			UDC_flip(Ser0UDCCS1, UDCCS1_SST);
			ep1_stall = 0;
		}
		
		if (!ep1_curdmalen) {
			UDC_flip(Ser0UDCCS1, UDCCS1_RPC);
			printk("usb_recv: RPC for non-existent buffer\n");
			return;
		}
		
		sa1100_dma_stop(dmachn_rx);

		if (status & UDCCS1_RPE) {
			//printk("usb_recv: error %x\n", status);
			UDC_flip(Ser0UDCCS1, UDCCS1_RPC);
			ep1_done(-EIO);
			return;
		}
		
		sa1100_dma_get_current(dmachn_rx, NULL, &dma_addr);
		pci_unmap_single(NULL, ep1_curdmapos, ep1_curdmalen, 
				 PCI_DMA_FROMDEVICE);
		len = dma_addr - ep1_curdmapos;
		if (len > ep1_curdmalen) {
			//printk("usb_recv: dma out of bound (%d/%d)\n", len, ep1_curdmalen);
			len = ep1_curdmalen;
		}
		if (len < ep1_curdmalen) {
			char *buf = ep1_curdmabuf + len;
			while (Ser0UDCCS1 & UDCCS1_RNE) {
				if (len >= ep1_curdmalen) {
					printk("usb_recv: too much data in fifo\n");
					break;
				}
				*buf++ = Ser0UDCDR;
				len++;
			}
		} else if (Ser0UDCCS1 & UDCCS1_RNE) {
			//printk("usb_recv: fifo screwed, shouldn't contain data\n");
			len = 0;
		}
		ep1_curdmalen = 0;  /* dma unmap already done */
		ep1_remain -= len;
		UDC_flip(Ser0UDCCS1, UDCCS1_RPC);
		ep1_done((len) ? 0 : -EIO);
	} else {
		//printk("usb_recv: UDCCS1 = %x\n", status);
		UDC_flip(Ser0UDCCS1, UDCCS1_SST);
		UDC_flip(Ser0UDCCS1, UDCCS1_RPC);
	}
}

int
sa1100_usb_recv(char *buf, int len, usb_callback_t callback)
{
	int flags;
	
	if (ep1_len)
		return -EBUSY;

	local_irq_save(flags);
	ep1_buf = buf;
	ep1_len = len;
	ep1_callback = callback;
	ep1_remain = len;
	ep1_curdmabuf = buf;
	ep1_curdmalen = 0;
	ep1_start();
	local_irq_restore(flags);
	
	return 0;
}

EXPORT_SYMBOL(sa1100_usb_recv);

void
sa1100_usb_recv_reset(void)
{
	ep1_reset();
}

EXPORT_SYMBOL(sa1100_usb_recv_reset);

