/*
 * Jornada 720 MCU keyboard/mouse driver
 *
 * 2000 by John Ankcorn, MIT LCS
 *
 * Changelog:
 *
 * This driver is based on linux/drivers/char/sa1111_keyb.c
 * Original declaration follows:
 *
 * linux/drivers/char/pc_keyb.c
 *
 * Separation of the PC low-level part by Geert Uytterhoeven, May 1997
 * See keyboard.c for the whole history.
 *
 * Major cleanup by Martin Mares, May 1997
 *
 * Combined the keyboard and PS/2 mouse handling into one file,
 * because they share the same hardware.
 * Johan Myreen <jem@iki.fi> 1998-10-08.
 *
 * Code fixes to handle mouse ACKs properly.
 * C. Scott Ananian <cananian@alumni.princeton.edu> 1999-01-29.
 *
 */

#include <linux/config.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/mm.h>
#include <linux/signal.h>
#include <linux/init.h>
#include <linux/kbd_ll.h>
#include <linux/delay.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/kbd_kern.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/pc_keyb.h>
#include <asm/keyboard.h>

#define MCU_GetBatteryData  0xc0
#define MCU_GetScanKeyCode	0x90
#define MCU_GetTouchSamples	0xa0
#define MCU_TxDummy         0x11
#define MCU_ErrorCode       0x00
#define SA1100_PPDR_LFCLK 	0x400

spinlock_t kbd_controller_lock = SPIN_LOCK_UNLOCKED;
/* used only by send_data - set by keyboard_interrupt */
static volatile unsigned char acknowledge = 0;
static volatile unsigned char resend = 0;
char mcu_invert[256];
static struct aux_queue *queue;        /* Mouse data buffer. */
static int aux_count = 0;

/*
 * Translation of escaped scancodes to keycodes.
 * This is now user-settable.
 * The keycodes 1-88,96-111,119 are fairly standard, and
 * should probably not be changed - changing might confuse X.
 * X also interprets scancode 0x5d (KEY_Begin).
 *
 * For 1-88 keycode equals scancode.
 */
#define SC_LIM 89

#define E0_KPENTER 96
#define E0_RCTRL   97
#define E0_KPSLASH 98
#define E0_PRSCR   99
#define E0_RALT    100
#define E0_BREAK   101  /* (control-pause) */
#define E0_HOME    102
#define E0_UP      103
#define E0_PGUP    104
#define E0_LEFT    105
#define E0_RIGHT   106
#define E0_END     107
#define E0_DOWN    108
#define E0_PGDN    109
#define E0_INS     110
#define E0_DEL     111
#define E0_YEN         124
#define E0_BACKSLASH   89
#define E1_PAUSE   119

static unsigned char high_keys[128 - SC_LIM] = {
  0, 0, 0, 0, 0, 0, 0,                   /* 0x59-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, 0,          /* 0x68-0x6f */
  0, 0, 0, 0, 0, 0, 0, 0,    /* 0x70-0x77 */
  0, 0, 0, 0,        /* 0x78-0x7b */
  0, 0, 0, 0                   /* 0x7c-0x7f */
};

#define E0_MACRO   112
#define E0_F13     113
#define E0_F14     114
#define E0_HELP    115
#define E0_DO      116
#define E0_F17     117
#define E0_KPMINPLUS 118
#define E0_OK  124
#define E0_MSLW        125
#define E0_MSRW        126
#define E0_MSTM        127

static unsigned char e0_keys[128] = {
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x00-0x07 */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x08-0x0f */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x10-0x17 */
  0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,            /* 0x18-0x1f */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x20-0x27 */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x28-0x2f */
  0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,            /* 0x30-0x37 */
  E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,              /* 0x38-0x3f */
  E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,              /* 0x40-0x47 */
  E0_UP, E0_PGUP, 0, E0_LEFT, E0_OK, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
  E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,              /* 0x50-0x57 */
  0, 0, 0, E0_MSLW, E0_MSRW, E0_MSTM, 0, 0,          /* 0x58-0x5f */
  0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x60-0x67 */
  0, 0, 0, 0, 0, 0, 0, E0_MACRO,                     /* 0x68-0x6f */
  //0, 0, 0, 0, 0, 0, 0, 0,                          /* 0x70-0x77 */
  0, 0, 0, 0, 0, E0_BACKSLASH, 0, 0,                         /* 0x70-0x77 */
  0, 0, 0, E0_YEN, 0, 0, 0, 0                        /* 0x78-0x7f */
};

void sa1111_leds(unsigned char leds)
{
/* dummy called from keyboard.c (keyb_leds() ) */
}

/* for "kbd-reset" cmdline param */
static int __init kbd_reset_setup(char *str)
{
       return 1;
}

__setup("kbd-reset", kbd_reset_setup);

int sa1111_setkeycode(unsigned int scancode, unsigned int keycode)
{
       if (scancode < SC_LIM || scancode > 255 || keycode > 127)
         return -EINVAL;
       if (scancode < 128)
         high_keys[scancode - SC_LIM] = keycode;
       else
         e0_keys[scancode - 128] = keycode;
       return 0;
}

int sa1111_getkeycode(unsigned int scancode)
{
       return
         (scancode < SC_LIM || scancode > 255) ? -EINVAL :
         (scancode < 128) ? high_keys[scancode - SC_LIM] :
           e0_keys[scancode - 128];
}

char sa1111_unexpected_up(unsigned char keycode)
{
   if (keycode >= SC_LIM || keycode == 85)
       return 0;
   else
       return 0200;
}

int sa1111_translate(unsigned char scancode, unsigned char *keycode,
                   char raw_mode)
{
       static int prev_scancode = 0;

       if (scancode == 0xe0 || scancode == 0xe1) {
               prev_scancode = scancode;
               return 0;
       }
       if (scancode == 0x00 || scancode == 0xff) {
               prev_scancode = 0;
               return 0;
       }
       scancode &= 0x7f;
       if (prev_scancode) {
         if (prev_scancode != 0xe0) {
             if (prev_scancode == 0xe1 && scancode == 0x1d) {
                 prev_scancode = 0x100;
                 return 0;
             } else if (prev_scancode == 0x100 && scancode == 0x45) {
                 *keycode = E1_PAUSE;
                 prev_scancode = 0;
             } else {
                 if (!raw_mode)
                   printk(KERN_INFO "keyboard: unknown e1 escape sequence\n");
                 prev_scancode = 0;
                 return 0;
             }
         } else {
             prev_scancode = 0;
             if (scancode == 0x2a || scancode == 0x36)
               return 0;
             if (e0_keys[scancode])
               *keycode = e0_keys[scancode];
             else {
                 if (!raw_mode)
                   printk(KERN_INFO "keyboard: unknown scancode e0 %02x\n",
                          scancode);
                 return 0;
             }
         }
       } else if (scancode >= SC_LIM) {
         *keycode = high_keys[scancode - SC_LIM];
         if (!*keycode) {
             if (!raw_mode) {
                 printk(KERN_INFO "keyboard: unrecognized scancode (%02x)"
                        " - ignored\n", scancode);
             }
             return 0;
         }
       } else
         *keycode = scancode;
       return 1;
}

static int mcu_byte(int arg_data)
{
int i;

	while ((Ser4SSSR & SSSR_TNF) == 0);
	i = 0;
	while ((GPLR & 0x400) && i++ < 400000)
		; /* wait for MCU */
	if (i >= 400000) {
		printk("mcu_byte: timed out\n");
        return -1;
    }
	Ser4SSDR = mcu_invert[arg_data] << 8;
	udelay(100);
	while ((Ser4SSSR & SSSR_RNE) == 0);
	i = Ser4SSDR;
	if (i > 0xff)
		printk("mcu_byte: read %x\n", i);
	return mcu_invert[ i & 0xff ];
}

static int mcu_read(void)
{
	return mcu_byte(MCU_TxDummy);
}

static void mcu_start(int arg_data)
{
int i;

	GPCR = 0x02000000; /* clear -> enable */
#if 0
	udelay(100);
#endif
	i = mcu_byte(arg_data);
	if (i != MCU_TxDummy)
		printk("mcu_start: [%x] bad start %x\n", arg_data, i);
}

static void mcu_end(void)
{
#if 0
	udelay(100);
#endif
	GPSR = 0x02000000; /* set */
}

static void keyboard_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
unsigned long flags;
unsigned int status = 0;
unsigned char scancode;

   kbd_pt_regs = regs;
   spin_lock_irqsave(&kbd_controller_lock, flags);
   while ((status & KBD_STAT_RXF) || (status & MSE_STAT_RXF)) {
       while (status & KBD_STAT_RXF) {
           scancode = KBDDATA & 0xff;
           handle_scancode(scancode, !(scancode & 0x80));
           tasklet_schedule(&keyboard_tasklet);
       }
       while (status & MSE_STAT_RXF) {
           scancode = MSEDATA & 0xff;
           add_mouse_randomness(scancode);
           if (aux_count) {
               int head = queue->head;
               queue->buf[head] = scancode;
               head = (head + 1) & (AUX_BUF_SIZE-1);
               if (head != queue->tail) {
                   queue->head = head;
                   if (queue->fasync)
                       kill_fasync(&queue->fasync, SIGIO, POLL_IN);
                   wake_up_interruptible(&queue->proc_list);
               }
           }
       }
   }
   spin_unlock_irqrestore(&kbd_controller_lock, flags);
}

static unsigned char get_from_queue(void)
{
       unsigned char result;
       unsigned long flags;

       spin_lock_irqsave(&kbd_controller_lock, flags);
       result = queue->buf[queue->tail];
       queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
       spin_unlock_irqrestore(&kbd_controller_lock, flags);
       return result;
} 

static inline int queue_empty(void)
{
       return queue->head == queue->tail;
}

static int fasync_aux(int fd, struct file *filp, int on)
{
       int retval;

       retval = fasync_helper(fd, filp, on, &queue->fasync);
       if (retval < 0)
               return retval;
       return 0;
} 

/*
 * Random magic cookie for the aux device
 */
#define AUX_DEV ((void *)queue)

static int release_aux(struct inode * inode, struct file * file)
{
       fasync_aux(-1, file, 0);
       if (--aux_count)
               return 0;
       // kbd_write_cmd(AUX_INTS_OFF);                     /* Disable controller ints */
       // kbd_write_command_w(KBD_CCMD_MOUSE_DISABLE);
       aux_free_irq(AUX_DEV);
       return 0;
}

/*
 * Install interrupt handler.
 * Enable auxiliary device.
 */
static int open_aux(struct inode * inode, struct file * file)
{
       if (aux_count++) {
               return 0;
       }
       mdelay(50);
       queue->head = queue->tail = 0;          /* Flush input queue */
       if (aux_request_irq(keyboard_interrupt, AUX_DEV)) {
               aux_count--;
               return -EBUSY;
       }
       /* aux_write_ack(AUX_ENABLE_DEV); / Enable aux device */
       return 0;
}

/*
 * Put bytes from input queue to buffer.
 */
static ssize_t read_aux(struct file * file, char * buffer,
                       size_t count, loff_t *ppos)
{
       DECLARE_WAITQUEUE(wait, current);
       ssize_t i = count;
       unsigned char c;

       if (queue_empty()) {
               if (file->f_flags & O_NONBLOCK)
                       return -EAGAIN;
               add_wait_queue(&queue->proc_list, &wait);
repeat:
               set_current_state(TASK_INTERRUPTIBLE);
               if (queue_empty() && !signal_pending(current)) {
                       schedule();
                       goto repeat;
               }
               current->state = TASK_RUNNING;
               remove_wait_queue(&queue->proc_list, &wait);
       }
       while (i > 0 && !queue_empty()) {
               c = get_from_queue();
               put_user(c, buffer++);
               i--;
       }
       if (count-i) {
               file->f_dentry->d_inode->i_atime = CURRENT_TIME;
               return count-i;
       }
       if (signal_pending(current))
               return -ERESTARTSYS;
       return 0;
}

/*
 * Write to the aux device.
 */
static ssize_t write_aux(struct file * file, const char * buffer,
                        size_t count, loff_t *ppos)
{
       ssize_t retval = 0;

       if (count) {
               ssize_t written = 0;
               if (count > 32)
                       count = 32; /* Limit to 32 bytes. */
               do {
                       char c;
                       get_user(c, buffer++);
                       printk ("[%s]: write_aux 0x%x\n", __FILE__, c);
                       written++;
               } while (--count);
               retval = -EIO;
               if (written) {
                       retval = written;
                       file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
               }
       }
       return retval;
}

static unsigned int aux_poll(struct file *file, poll_table * wait)
{
       poll_wait(file, &queue->proc_list, wait);
       if (!queue_empty())
               return POLLIN | POLLRDNORM;
       return 0;
}

struct file_operations psaux_fops = {
       read:           read_aux,
       write:          write_aux,
       poll:           aux_poll,
       open:           open_aux,
       release:        release_aux,
       fasync:         fasync_aux,
};

/*
 * Initialize driver.
 */
static struct miscdevice psaux_mouse = {
       PSMOUSE_MINOR, "psaux", &psaux_fops
};
int jcacount;
long jcail;
void jcabozo(void)
{
jcacount++;
}

void __init sa1111_init_hw(void)
{
int i, this_count;
int x1, x2, x3, y1, y2, y3;

    kbd_request_region();
    queue = (struct aux_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
    memset(queue, 0, sizeof(*queue));
    queue->head = queue->tail = 0;
    init_waitqueue_head(&queue->proc_list);

	PPSR &= ~(SA1100_PPDR_LFCLK | 0x80 | 0x10);
	PPDR |= SA1100_PPDR_LFCLK | 0x80 | 0x10;
	printk ("[%s]: sa1111_init_hw entry\n", __FILE__);
	/* Take the MCU out of reset mode */
	PPSR |= SA1100_PPDR_LFCLK;
	printk ("now wait after reset done\n");
	for (i = 0; i <= 255; i++)
		mcu_invert[i] = ( ((0x80 & i) >> 7) | ((0x40 & i) >> 5)
		  | ((0x20 & i) >> 3) | ((0x10 & i) >> 1) | ((0x08 & i) << 1)
		  | ((0x04 & i) << 3) | ((0x02 & i) << 5) | ((0x01 & i) << 7) );

	GPSR = 0x02000000; /* set */
	GPDR |= 0x02000000; /* GPIO(25) low -> enable MCU */
	Ser4SSCR0 = 0x0307;
	Ser4MCCR0 = 0;
	Ser4SSSR = 0;	/* remove any rcv overrun errors */
	Ser4SSCR1 = 0x18;
	Ser4SSCR0 = 0x0387;

	/* drain any data already there */
	while (Ser4SSSR & SSSR_RNE)  i = Ser4SSDR;

#if 0
	mcu_start(MCU_GetBatteryData);
	printk("get2 bat %x\n", mcu_read());
	printk("get3 bat %x\n", mcu_read());
	printk("get4 bat %x\n", mcu_read());
	mcu_end();

for (jcail = 0; jcail < 20000000; jcail++) jcabozo();
printk("after loop\n");
#endif
    /* Flush any pending input. */
	mcu_start(MCU_GetScanKeyCode);
	this_count = mcu_read();
	printk("get key count %x\n", this_count);
	while (this_count-- > 0) {
		printk("get key %x\n", mcu_read());
	}
	mcu_end();

#if 0
	mcu_start(MCU_GetTouchSamples);
	x1 = mcu_read();
	x2 = mcu_read();
	x3 = mcu_read();
	i = mcu_read();
	x1 |= (i & 3) << 8;
	x2 |= (i & 0xc) << 6;
	x3 |= (i & 0x30) << 4;
	y1 = mcu_read();
	y2 = mcu_read();
	y3 = mcu_read();
	i = mcu_read();
	y1 |= (i & 3) << 8;
	y2 |= (i & 0xc) << 6;
	y3 |= (i & 0x30) << 4;
	mcu_end();
	printk("x %x %x %x y %x %x %x\n", x1, x2, x3, y1, y2, y3);
#endif

    misc_register(&psaux_mouse);
    // kbd_write_command_w(KBD_CCMD_MOUSE_ENABLE); /* Enable Aux. */
    /* Ok, finally allocate the IRQ, and off we go.. */
    kbd_request_irq(keyboard_interrupt);
}
